"""This module contains commonly-used Op implementations."""

from __future__ import annotations

import asyncio
import functools
import inspect
import logging
from collections.abc import Awaitable, Callable
from functools import lru_cache
from typing import TYPE_CHECKING, Generic, TypeVar, cast, overload

import numpy as np
import tenacity
from aviary.core import Message, Tool, ToolRequestMessage, is_coroutine_callable
from lmi import (
    EmbeddingModel,
    HybridEmbeddingModel,
    LiteLLMEmbeddingModel,
    LLMResult,
    SparseEmbeddingModel,
)
from lmi import LiteLLMModel as LLMModel
from pydantic import BaseModel

from .memory import Memory, MemoryModel, UIndexMemoryModel
from .op_utils import CallID, _lazy_import_tree, get_call_id, get_training_mode
from .ops import GradInType, Op, OpCtx, ResultOrValue, TOutput_co

if TYPE_CHECKING:
    import tree

logger = logging.getLogger(__name__)


def logsumexp(a: np.ndarray | list[float]) -> float:
    a_max = np.max(a)
    return a_max + np.log(np.sum(np.exp(a - a_max)))


TOutput = TypeVar("TOutput")


class IdentityOp(Op[TOutput_co]):
    """
    An operation that simply returns the input value.

    NOTE: this op is equivalent to FxnOp(lambda x: x).
    """

    # Can't have covariant TypeVar as a parameter type
    async def forward(self, value: TOutput) -> TOutput:
        # We assume value already has the correct run_id from its producer
        return value

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return [], {"value": grad_output}


class StopGradOp(IdentityOp[TOutput_co]):
    """Pass through Op that terminates gradients in the backward pass."""

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        from .gradient_estimators import assign_constant_grads

        return assign_constant_grads(input_args, input_kwargs, None)


TConfig = TypeVar("TConfig", bound=BaseModel | dict)


class ConfigOp(Op[TConfig], Generic[TConfig]):
    """An operation that contains a configuration object."""

    def __init__(self, config: TConfig):
        self.config = config

    async def forward(self) -> TConfig:
        return self.config

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        # Check that the grad_output structure is consistent with our config
        _lazy_import_tree().assert_same_structure(
            grad_output, ctx.get(call_id, "output").value, check_types=False
        )

        # Terminate here - we're a leaf since a ConfigOp takes no inputs
        return [], {}


TResult = TypeVar("TResult")


class Cacheable(Generic[TResult]):
    def __init__(self, co: Awaitable[TResult]) -> None:
        self.co = co
        self.done = False
        self.result: TResult | None = None
        self.lock = asyncio.Lock()

    async def get_result(self) -> TResult | None:
        async with self.lock:
            if not self.done:
                self.result = await self.co
                self.done = True
            return self.result

    def __await__(self):
        return self.get_result().__await__()


def async_cache(func):
    @functools.lru_cache(maxsize=1024)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return Cacheable(co=func(*args, **kwargs))

    return wrapper


class FxnOp(Op[TOutput_co]):
    """
    Wrap a function for a straight through gradient approximation for all args/kwargs.

    Basically, consider the fxn as a transform upon the inputs during the forward pass,
    and propagating the same gradient for all inputs during the backward pass.
    """

    def __init__(
        self,
        fxn: Callable[..., TOutput_co] | Callable[..., Awaitable[TOutput_co]],
        cache: bool = False,
        fxn_name: str | None = None,  # useful for lambdas
    ):
        if cache:
            self.fxn = (
                async_cache(fxn) if is_coroutine_callable(fxn) else lru_cache()(fxn)
            )
        else:
            self.fxn = fxn

        try:
            self.fxn_name = fxn_name or fxn.__name__
        except AttributeError:  # unittest.mock.Mock or lambda
            self.fxn_name = str(fxn)

        # override forward args with the signature of the function
        fwd_sig = inspect.signature(self.fxn)
        self._fwd_args = list(fwd_sig.parameters.values())

    def __repr__(self) -> str:
        return f"{type(self).__name__} {self.fxn_name} ({id(self)})"

    async def forward(self, *args, **kwargs) -> TOutput_co:
        if is_coroutine_callable(self.fxn):
            return await self.fxn(*args, **kwargs)
        return self.fxn(*args, **kwargs)

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        from .gradient_estimators import assign_constant_grads

        return assign_constant_grads(input_args, input_kwargs, 0.0)


class PromptOp(FxnOp[str]):
    """An operation that formats kwargs into a prompt string."""

    async def _fxn(
        self, prompt_kwargs: dict[str, str] | None = None, **kwargs: str
    ) -> str:
        return self.prompt.format(**{**(prompt_kwargs or {}), **kwargs})

    def __init__(self, prompt: str):
        self.prompt = prompt
        super().__init__(fxn=self._fxn, cache=False)

    def __repr__(self) -> str:
        # we want to use Op.__repr__, not FxnOp.__repr__
        return super(FxnOp, self).__repr__()


class ResponseValidationError(Exception):
    """Raised when a response from the LLM does not pass user-specified validator."""


class LLMCallOp(Op[Message]):
    """An operation for LLM calls interaction."""

    def __init__(
        self,
        num_samples_logprob_estimate: int = 0,
        response_validator: Callable[[LLMResult], Awaitable[None] | None] | None = None,
    ) -> None:
        """Initializes the LLMCallOp.

        Args:
            num_samples_logprob_estimate: The number of samples used to estimate the partition
                function at T!=1. Defaults to 0 (calculation is skipped).
            response_validator: An optional callable (can be async) that validates the response.
                It should raise an exception if the response is invalid. The Op will retry up
                to `config.get('num_retries', 0)` times if validation fails.
        """
        super().__init__()
        self.num_samples_partition_estimate = num_samples_logprob_estimate
        self.response_validator = response_validator

    @overload
    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: list[Tool] = ...,
        tool_choice: Tool | str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> ToolRequestMessage: ...

    @overload
    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: None = None,
        tool_choice: str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> Message: ...

    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: list[Tool] | None = None,
        tool_choice: Tool | str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> Message:
        """Calls the LLM.

        Args:
            config: Configuration passed to LLMModel.
            msgs: Input messages to prompt model with.
            tools: A list of Tools that the model may call, if supported.
            tool_choice: Configures how the model should choose a tool.
                Can be a Tool or a string; see here for string options:
                https://platform.openai.com/docs/guides/function-calling#configuring-function-calling-behavior-using-the-tool_choice-parameter
                NOTE: if `tools` is None or empty, this parameter is ignored.

        Returns:
            Output message from the model.
        """
        model = LLMModel(config=config)

        if not tools:
            # if no tools are provided, tool_choice must be 'none'
            tool_choice = "none"

        result = await self._call_single_and_maybe_validate(
            model=model,
            messages=msgs,
            tools=tools,
            tool_choice=tool_choice,
            # Since config is shared between our LLMModel and our `call` options
            # we need to ensure we remove keys which work with LLMModel
            # but not `call`.
            **{k: v for k, v in config.items() if k != "router_kwargs"},
        )
        if result.messages is None:
            raise ValueError("No messages returned")

        # if not set, assume temp = 1. TODO: when would it not be set?
        temperature: float = config.get("temperature", 1.0)

        # Compute a Monte Carlo estimate of the logprob of this sequence at the given temperature.
        logprob = await self.compute_logprob(
            raw_log_p=result.logprob,
            temperature=temperature,
            model=model,
            messages=msgs,
            tools=tools,
            tool_choice=tool_choice,
        )

        call_id = get_call_id()
        self.ctx.update(call_id, "result", result)
        # This is the logprob of this sequence according to the raw model, without
        # any temperature/top-p distribution shaping.
        self.ctx.update(call_id, "raw_logprob", result.logprob)

        self.ctx.update(call_id, "temperature", temperature)
        self.ctx.update(call_id, "logprob", logprob)

        return result.messages[0]

    async def _call_single_and_maybe_validate(
        self, model: LLMModel, num_retries: int = 0, **kwargs
    ) -> LLMResult:
        if not self.response_validator:
            # If a response validator is not supplied, then we should not do any retries here - leave
            # that for LiteLLM to handle.
            return await model.call_single(**kwargs)

        # NOTE: `num_retries` also gets passed to LiteLLM, so there could a maximum of
        # `num_retries**2` retries. TODO: consider if we should have separate parameters
        # for LiteLLM and validation retries.
        @tenacity.retry(
            retry=tenacity.retry_if_exception_type(ResponseValidationError),
            # num_retries+1 because the first call is not a retry
            stop=tenacity.stop_after_attempt(num_retries + 1),
            wait=tenacity.wait_fixed(1),
        )
        async def call_and_validate() -> LLMResult:
            result = await model.call_single(**kwargs)

            try:
                validated = cast("Callable", self.response_validator)(result)
                if inspect.isawaitable(validated):
                    validated = await validated
            except Exception as e:
                raise ResponseValidationError(
                    f"Response validator failed: {self.response_validator!r}"
                ) from e

            return result

        return await call_and_validate()

    async def compute_logprob(
        self,
        raw_log_p: float | None,
        temperature: float,
        model: LLMModel,
        **model_kwargs,
    ) -> float | None:
        """This method computes a Monte Carlo estimate of logprob for a given temperature.

        It takes as input the logprob at T=1. The derivation is in Section 5.1 of the Aviary notes.
        """
        if temperature == 1:
            return raw_log_p

        if temperature == 0:
            return 1.0

        if raw_log_p is None or self.num_samples_partition_estimate == 0:
            return None

        # TODO: possibly move to MultipleCompletionLLMModel here, though we need to check that the estimates
        # are consistent - not sure we'd be sampling from the same distribution as N independent samples.
        # TODO: think about whether sampling params besides temperature need to be accounted for, like top_p
        results = await asyncio.gather(*[
            model.call_single(temperature=1, **model_kwargs)
            for _ in range(self.num_samples_partition_estimate)
        ])
        temp_factor = 1.0 / temperature - 1.0

        # Partition function estimate:
        # Z_T = E_P[ e^(lnP/T - lnP) ]
        log_Z_T = logsumexp([
            temp_factor * cast("float", result.logprob) for result in results
        ]) - np.log(self.num_samples_partition_estimate)

        return (raw_log_p / temperature) - log_Z_T

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        # By default, we want to descend into config, but not msgs/tools/tool_choice
        # Essentially: we can think of each config field as an independent parameter,
        # but not necessarily each message or tool.

        # tree.map_structure allows us to assign a gradient of 0 to all fields of config
        grad_config = _lazy_import_tree().map_structure(
            lambda _: 0.0, input_kwargs["config"]
        )
        grad_kwargs = {"config": grad_config}
        for arg in ("msgs", "tools", "tool_choice"):
            if arg in input_kwargs:
                grad_kwargs[arg] = 0.0

        return [], grad_kwargs

    def get_examples(self) -> list[tuple[LLMResult, float]]:
        examples = [
            (
                self.ctx.get(c, "result", None),
                # get 'model' kwarg from grad_input
                # use default of None if not found
                self.ctx.get(c, "grad_input", default=([], {}))[1].get("model", None),
            )
            for c in self.get_call_ids()
        ]
        # filter out the None values
        return [(e, w) for e, w in examples if e is not None]

    @staticmethod
    def anthropic_response_validator(response: LLMResult) -> None:
        """Sometimes Anthropic models respond with garbled tool calls.

        Specifically, parameters get injected into the tool name. This obviously
        breaks the tool call, but also messes up the subsequent tool response message.
        So instead, check here and force a retry if configured.
        """
        if response.messages is None:
            return

        for msg in response.messages:
            if not isinstance(msg, ToolRequestMessage):
                continue

            for tc in msg.tool_calls:
                if any(x in tc.function.name for x in "{}<>()"):
                    err_msg = f"Found bad tool call: {tc}"
                    logger.error(err_msg)
                    raise ValueError(err_msg)


class MemoryOp(Op[list[Memory]]):
    """An operation for managing memory retrieval and storage."""

    def __init__(self, memory_model: MemoryModel | None = None):
        super().__init__()
        self.memory_model = memory_model or UIndexMemoryModel(
            embedding_model=EmbeddingModel.from_name("sparse")
        )

    async def forward(
        self,
        query: str,
        input: str | None = None,  # noqa: A002
        matches: int = 3,
    ) -> list[Memory]:
        """Retrieve relevant memories based on a query."""
        if get_training_mode():
            call_id = get_call_id()
            self.ctx.update(call_id, "query", query)
            self.ctx.update(call_id, "memory_input", input)
        return await self.memory_model.get_memory(query, matches)

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        """Backward pass for memory retrieval - goes back to item."""
        from .gradient_estimators import assign_constant_grads

        return assign_constant_grads(input_args, input_kwargs, 0.0)


class EmbeddingOp(Op):
    """A general operation for embedding text using LiteLLM."""

    def __init__(
        self,
        *,
        dense_embedding: str = "text-embedding-3-small",
        dense_embedding_dim: int = 512,
        sparse_embedding_dim: int = 0,
        **embedding_model_kwargs,
    ):
        if "timeout" not in embedding_model_kwargs:
            embedding_model_kwargs.setdefault("timeout", 60)
        emb_models: list[EmbeddingModel] = []
        if dense_embedding_dim > 0:
            emb_models.append(
                LiteLLMEmbeddingModel(
                    name=dense_embedding,
                    ndim=dense_embedding_dim,
                    embed_kwargs=embedding_model_kwargs,
                )
            )
        if sparse_embedding_dim > 0:
            emb_models.append(SparseEmbeddingModel(ndim=sparse_embedding_dim))
        self.embedding = HybridEmbeddingModel(models=emb_models)

    async def forward(self, string_input: str) -> np.ndarray:
        return np.array(await self.embedding.embed_document(string_input))

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return [], {"string_input": None}



import asyncio
import json
import logging
import os
import time
from collections import defaultdict
from collections.abc import Callable, Collection, Iterable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import aiofiles
from aviary.core import (
    Environment,
    Message,
    MessagesAdapter,
    TaskDataset,
    Tool,
    ToolRequestMessage,
)

from ldp.agent import Agent
from ldp.data_structures import Trajectory, Transition
from ldp.graph import OpCtx, OpResult

try:
    import wandb
except ImportError:
    wandb = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from ldp.alg.optimizer.replay_buffers import ReplayBuffer

logger = logging.getLogger(__name__)


class Callback:
    """Base class for callbacks used by RolloutManager/Evaluator/OnlineTrainer.

    Pseudocode to demonstrate how callback methods are invoked (marked as *):

    RolloutManager.sample_trajectories():
        callback.before_rollout() *
        env.reset()
        callback.after_env_reset() *
        agent.init_state()
        callback.after_agent_init_state() *
        while not done:
            callback.before_transition() *
            agent.get_asv()
            callback.after_agent_get_asv() *
            env.step()
            callback.after_env_step() *
            callback.after_transition() *

    Evaluator.evaluate / OnlineTrainer._eval_loop():
        callback.before_eval_loop() *
        for batch in eval_dataset:
            rollout_manager.sample_trajectories()
            callback.after_eval_step() *
        callback.after_eval_loop() *

    OfflineTrainer / OnlineTrainer.train():
        for batch in train_dataset:
            rollout_manager.sample_trajectories() # if online
            optimizer.aggregate()
            if updating_optimizer:
                optimizer.update()
                callback.after_update() *
            callback.after_train_step() *
    """

    async def before_rollout(self, traj_id: str, env: Environment) -> None:
        """Invoked by runners when each rollout starts (at trajectory start)."""

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        """Invoked by runners before each transition and after agent and env reset."""

    async def after_agent_init_state(self, traj_id: str, init_state: Any) -> None:
        """Invoked by runners after agent.init_state()."""

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        """Invoked by runners after agent.get_asv()."""

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        """Invoked by runners after env.reset()."""

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        """Invoked by runners after env.step()."""

    async def after_transition(
        self, traj_id: str, agent: Agent, env: Environment, transition: Transition
    ) -> None:
        """Invoked by runners after each transition."""

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Invoked by OnlineTrainer after each training step."""

    async def before_eval_loop(self) -> None:
        """Invoked by Evaluator and OnlineTrainer before the evaluation loop."""

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Invoked by Evaluator and OnlineTrainer after each evaluation step."""

    async def after_eval_loop(self) -> None:
        """Invoked by Evaluator and OnlineTrainer after the evaluation loop."""

    async def after_update(self) -> None:
        """Invoked by OnlineTrainer after each optimizer.update() call."""


class StoreTrajectoriesCallback(Callback):
    """Simple callback that stores train/eval trajectories in an in-memory list."""

    def __init__(self):
        self.train_trajectories = []
        self.eval_trajectories = []

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        self.train_trajectories.extend(trajectories)

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        self.eval_trajectories.extend(trajectories)


class TrajectoryFileCallback(Callback):
    """Callback that writes trajectories to a file."""

    @staticmethod
    def default_serialize_env(env: Environment, transition: Transition) -> str | None:
        """Export a JSON-serialized Frame if the transition is done."""
        if not transition.done:
            return None
        try:
            frame = env.export_frame()
        except NotImplementedError:
            return None  # Allow for envs that didn't implement export_frame()
        return frame.model_dump_json(exclude={"state"}, indent=2)

    def __init__(
        self,
        output_dir: str | os.PathLike,
        serialize_env: Callable[
            [Environment, Transition], str | None
        ] = default_serialize_env,
    ):
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.traj_files: dict[str, Path] = {}
        self.env_files: dict[str, Path] = {}
        self.trajs: dict[str, Trajectory] = defaultdict(Trajectory)

        self.serialize_env = serialize_env

    def _make_traj_filename(self, traj_id: str, env: Environment) -> str:
        """Create the filename for the output file."""
        return f"{traj_id}.jsonl"

    async def after_transition(
        self, traj_id: str, agent: Agent, env: Environment, transition: Transition
    ) -> None:
        if traj_id not in self.traj_files:
            self.traj_files[traj_id] = self.output_dir / self._make_traj_filename(
                traj_id, env
            )
            self.env_files[traj_id] = self.output_dir / f"{traj_id}_env.json"

        traj = self.trajs[traj_id]
        traj.steps.append(transition)

        async def possibly_dump_env() -> None:
            async with aiofiles.open(self.env_files[traj_id], "w") as f:
                env_or_none = self.serialize_env(env, transition)
                if env_or_none is not None:
                    await f.write(env_or_none)

        await asyncio.gather(
            possibly_dump_env(), traj.to_jsonl(self.traj_files[traj_id])
        )


class StoreEnvironmentsCallback(Callback):
    """Callback to store the environment underlying each trajectory."""

    def __init__(self):
        self.traj_id_to_envs: dict[str, Environment] = {}

    async def before_rollout(self, traj_id: str, env: Environment) -> None:
        self.traj_id_to_envs[traj_id] = env


class RolloutDebugDumpCallback(Callback):
    """Dump JSONL files for each agent and environment step to a directory."""

    def __init__(self, output_dir: os.PathLike | str):
        """Initialize.

        Args:
            output_dir: Directory to place JSONL files.
        """
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.out_files: dict[str, Path] = {}

    def _get_out_file(self, traj_id: str) -> Path:
        if traj_id not in self.out_files:
            self.out_files[traj_id] = self.output_dir / f"{traj_id}.jsonl"
        return self.out_files[traj_id]

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state,
        obs: list[Message],
    ) -> None:
        self.start = time.time()

    def _get_elapsed_time(self, reset: bool = True) -> float:
        elapsed = time.time() - self.start
        if reset:
            self.start = time.time()
        return elapsed

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        log_jsonl = json.dumps({
            "event": "AGENT_GET_ASV",
            "elapsed": self._get_elapsed_time(),
            "action": action.value.model_dump(),
            "value": value,
        })
        async with aiofiles.open(self._get_out_file(traj_id), "a") as f:
            await f.write(log_jsonl + "\n")

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        log_jsonl = json.dumps({
            "event": "ENV_STEP",
            "elapsed": self._get_elapsed_time(),
            "obs": MessagesAdapter.dump_python(obs),
            "reward": reward,
            "done": done,
            "truncated": trunc,
        })
        async with aiofiles.open(self._get_out_file(traj_id), "a") as f:
            await f.write(log_jsonl + "\n")


class ComputeTrajectoryMetricsMixin:
    """Mixin for TaskDataset classes to enable them to compute metrics."""

    # Tools or tool names to include in trajectory metrics
    tools_to_track: Collection[str | Tool] = set()

    def compute_trajectory_metrics(
        self,
        trajectories: Sequence[Trajectory],
    ) -> dict[str, list[float]]:
        metrics: dict[str, list[float]] = {
            "reward": [
                sum(step.reward for step in traj.steps) for traj in trajectories
            ],
            "truncation_rate": [
                sum(step.truncated for step in traj.steps) for traj in trajectories
            ],
            "avg_value": [
                sum(step.value for step in traj.steps) / len(traj.steps)
                for traj in trajectories
            ],
            "num_steps": [len(traj.steps) for traj in trajectories],
            "failures": [traj.failed for traj in trajectories],
        }
        for tool in self.tools_to_track:  # Default of empty set means this is not run
            if isinstance(tool, Tool):
                tool = tool.info.name
            metrics[f"tool_{tool}"] = [
                sum(
                    sum(tc.function.name == tool for tc in s.action.value.tool_calls)
                    for s in traj.steps
                    if isinstance(s.action, OpResult)
                )
                for traj in trajectories
            ]
        return metrics


class TrajectoryMetricsCallback(Callback):
    """
    Compute metrics that are defined by task datasets.

    NOTE: evaluation portion's after_eval_step/loop() is not concurrency safe because
    trajectories should be stored in the order of after_eval_step() calls.
    """

    def __init__(
        self,
        train_dataset: TaskDataset | None = None,
        eval_dataset: TaskDataset | None = None,
        track_tool_usage: bool = False,
    ):
        self._datasets = train_dataset, eval_dataset
        self._track_tool_usage = track_tool_usage
        for ds in self._datasets:
            if ds and not isinstance(ds, ComputeTrajectoryMetricsMixin):
                raise ValueError(
                    f"Dataset {ds} didn't implement"
                    f" {ComputeTrajectoryMetricsMixin.__name__}, which is required for"
                    " this callback."
                )
        self._train_metrics_fn = (
            train_dataset.compute_trajectory_metrics if train_dataset else None  # type: ignore[attr-defined]
        )
        self._eval_metrics_fn = (
            eval_dataset.compute_trajectory_metrics if eval_dataset else None  # type: ignore[attr-defined]
        )

        self._train_metrics: dict[str, list[float]] | None = None
        self._eval_metrics: dict[str, list[float]] = {}

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        for ds in (ds for ds in self._datasets if ds):
            if self._track_tool_usage:
                cast("ComputeTrajectoryMetricsMixin", ds).tools_to_track = {
                    t.info.name for t in tools
                }

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        if self._train_metrics_fn is not None:
            self._train_metrics = self._train_metrics_fn(trajectories)

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        if self._eval_metrics_fn is not None:
            for k, v in self._eval_metrics_fn(trajectories).items():
                if k not in self._eval_metrics:
                    # Don't use defaultdict - error prone in user code
                    self._eval_metrics[k] = []
                self._eval_metrics[k].extend(v)

    async def after_eval_loop(self) -> None:
        self._eval_metrics.clear()


class MeanMetricsCallback(TrajectoryMetricsCallback):
    """Take a mean of all metrics."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._train_means: dict[str, float] | None = None
        self._eval_means: dict[str, float] | None = None

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        await super().after_train_step(trajectories)
        if self._train_metrics is not None:
            # may be None if train_dataset was not provided
            self._train_means = self._compute_means(self._train_metrics)

    async def after_eval_loop(self) -> None:
        if self._eval_metrics:
            # may be empty if eval_dataset was not provided
            self._eval_means = self._compute_means(self._eval_metrics)
        await super().after_eval_loop()

    @staticmethod
    def _compute_means(metrics: dict[str, list[float]]) -> dict[str, float]:
        return {k: sum(v) / len(v) for k, v in metrics.items()}

    @property
    def train_means(self) -> dict[str, float]:
        if self._train_means is None:
            raise RuntimeError(
                "Training means are only available after this callback is invoked."
            )
        return self._train_means

    @property
    def eval_means(self) -> dict[str, float]:
        if self._eval_means is None:
            raise RuntimeError(
                "Evaluation means are only available after this callback is invoked."
            )
        return self._eval_means


class WandBLoggingCallback(TrajectoryMetricsCallback):
    def __init__(self, *args, **kwargs):
        if wandb is None:
            raise ImportError(
                f"{type(self).__name__} processing requires the 'monitor' extra for"
                " 'wandb'. Please: `pip install aviary-internal[monitor]`."
            )
        super().__init__(*args, **kwargs)

        self._num_train_step = 0

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        await super().after_train_step(trajectories)
        self._num_train_step += 1

        if self._train_metrics is None:
            return

        # Each wandb.log() increments the wandb step by 1. Log the training step here
        # so we can use it as an x-axis for training metrics that are logged by different
        # wandb.log() calls.
        wandb.log(
            {
                f"train/{key}_mean": sum(vals) / len(vals)
                for key, vals in self._train_metrics.items()
            }
            | {"train/step": self._num_train_step}
        )

    async def after_eval_loop(self) -> None:
        if not self._eval_metrics:
            return

        wandb.log({
            f"eval/{key}_mean": sum(vals) / len(vals) if vals else None
            for key, vals in self._eval_metrics.items()
        })

        await super().after_eval_loop()


class ClearContextCallback(Callback):
    def __init__(self, op_names: Iterable[str] | None = None):
        self._op_names = op_names

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        OpCtx.clear_contexts(self._op_names)

    async def after_update(self) -> None:
        OpCtx.clear_contexts(self._op_names)


class LoggingCallback(MeanMetricsCallback):
    """Custom callback for logging filtered metrics (e.g., pass rates) to the console.

    This callback extends the `MeanMetricsCallback` and allows logging of user-specified metrics
    after each training step and after the evaluation loop. It calculates the specified metrics
    (e.g., pass rates) from the trajectories and logs the results.
    """

    def __init__(
        self,
        train_dataset: TaskDataset | None = None,
        eval_dataset: TaskDataset | None = None,
        metrics_to_log: Collection[str] | None = None,
    ):
        """Initialize the callback with a list of metric keys to log.

        Args:
            train_dataset: The training dataset for computing metrics.
            eval_dataset: The evaluation dataset for computing metrics.
            metrics_to_log: Optional metric names (e.g., ["pass"]) to log.
                            If left as default of None, all metrics will be logged.
        """
        super().__init__(train_dataset, eval_dataset)
        self.metrics_to_log = (
            metrics_to_log or set()
        )  # If no metrics provided, log all by default

    def _log_filtered_metrics(self, metrics: dict[str, float], step_type: str) -> None:
        """Helper function to log only the specified metrics.

        Args:
            metrics: Dictionary of calculated means for the current step (e.g., train or eval).
            step_type: The type of step (e.g., "Train" or "Eval") for logging purposes.
        """
        if self.metrics_to_log:
            for metric in self.metrics_to_log:
                if metric in metrics:
                    logger.info(
                        f"{metric.upper()} RATE ({step_type}): {metrics[metric]:.5f}"
                    )
        else:
            # Log all metrics if no specific ones are provided
            logger.info(f"{step_type} Metrics: {metrics}")

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Log metrics and pass rate after each training step.

        This method is called after every training step, calculating and logging
        the training metrics and pass rate.

        Args:
            trajectories: A sequence of trajectories from the training step.
        """
        await super().after_train_step(trajectories)  # Call the parent to compute means
        if self.train_means:
            self._log_filtered_metrics(self.train_means, step_type="Train")

    async def after_eval_loop(self) -> None:
        """Log metrics and pass rate after the evaluation loop.

        This method is called after the evaluation loop finishes, calculating and logging
        the evaluation metrics and pass rate.
        """
        await super().after_eval_loop()  # Call the parent to compute means
        if self.eval_means:
            self._log_filtered_metrics(self.eval_means, step_type="Eval")


class TerminalPrintingCallback(Callback):
    """Callback that prints action, observation, and timing information to the terminal."""

    def __init__(self):
        self.start_time = None
        # try now, rather than start running and die
        try:
            from rich.pretty import pprint  # noqa: F401
        except ImportError as e:
            raise ImportError(
                f"rich is required for {type(self).__name__}. Please install it with"
                " `pip install rich`."
            ) from e

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        """Start the timer before each transition."""
        self.start_time = time.time()

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        from rich.pretty import pprint

        print("\nAction:")
        pprint(action.value, expand_all=True)

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        from rich.pretty import pprint

        # Compute elapsed time
        if self.start_time is not None:
            elapsed_time = time.time() - self.start_time
            self.start_time = None  # Reset timer
        else:
            elapsed_time = 0.0
        print("\nObservation:")
        pprint(obs, expand_all=True)
        print(f"Elapsed time: {elapsed_time:.2f} seconds")


class ClearOptimizerBuffersCallback(Callback):
    """Invoke the clear method on buffer(s) after each optimizer update."""

    def __init__(self, *buffers: "ReplayBuffer"):
        self._buffers = list(buffers)

    async def after_update(self) -> None:
        for b in self._buffers:
            b.clear()



import itertools
import random
import tempfile
from copy import deepcopy
from typing import Any, cast
from unittest.mock import AsyncMock

import pytest
from aviary.core import Environment, Frame, Message, Tool, ToolRequestMessage
from litellm import AuthenticationError
from pydantic import BaseModel

from ldp.agent import Agent, SimpleAgent, SimpleAgentState
from ldp.alg import (
    BeamSearchRollout,
    Callback,
    RolloutManager,
    StoreEnvironmentsCallback,
    TreeSearchRollout,
)
from ldp.data_structures import Trajectory, Transition
from ldp.graph import FxnOp, OpResult, compute_graph, set_training_mode


class DummyEnv(Environment[None]):
    def __init__(self, instance_id: int | None = None):
        self.tools = [Tool.from_function(self.talk)]
        self._instance_id = instance_id
        self.close_mock = AsyncMock()

    async def get_id(self) -> str:
        if self._instance_id is None:
            raise ValueError("No instance ID was configured.")
        return str(self._instance_id)

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        return [Message(content="Hello!")], self.tools

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        if action.tool_calls:
            responses = cast("list[Message]", await self.exec_tool_calls(action))
        else:
            responses = [Message(content="Use the 'talk' tool to speak.")]

        return responses, 0.0, False, False

    async def talk(self, message: str) -> str:
        """Say something to me.

        Args:
            message (str): what you want to say

        Returns:
            str: my response
        """
        return message

    def export_frame(self) -> Frame:
        return Frame()

    async def close(self) -> None:
        await self.close_mock()


async def count_exclamations(traj: Trajectory) -> float:  # noqa: RUF029
    last_step = traj.steps[-1]
    agent_state = cast("SimpleAgentState", last_step.next_agent_state)
    return float(
        sum(m.content.count("!") for m in agent_state.messages if m.content is not None)
    )


@pytest.mark.vcr
@pytest.mark.parametrize("training", [True, False])
@pytest.mark.asyncio
async def test_rollout(training: bool) -> None:
    agent = SimpleAgent()
    counter_callback = DummyCallback()
    env_storage_callback = StoreEnvironmentsCallback()
    set_training_mode(training)
    rollout_manager = RolloutManager(
        agent,
        catch_agent_failures=False,
        catch_env_failures=False,
        callbacks=[counter_callback, env_storage_callback],
    )
    envs = [DummyEnv(instance_id=1), DummyEnv()]
    trajs = await rollout_manager.sample_trajectories(environments=envs, max_steps=1)
    first_traj, second_traj = trajs
    assert first_traj.traj_id
    assert (
        first_traj.metadata.get("env_id")
        == "1"
        == await env_storage_callback.traj_id_to_envs[first_traj.traj_id].get_id()
    )
    assert second_traj.metadata.get("env_id") is None

    for env in envs:
        env.close_mock.assert_awaited_once_with()

    assert all(
        tx.metadata.get(f"time_elapsed_{fn}") is not None
        for fn in (
            "before_transition",
            "agent_get_asv",
            "after_agent_get_asv",
            "env_step",
            "after_env_step",
        )
        for traj in trajs
        for tx in traj.steps
    ), "All transitions should have timing metadata"

    # Let's check we can serialize and deserialize the trajectories
    for traj in trajs:
        assert traj.traj_id
        assert traj.traj_id in env_storage_callback.traj_id_to_envs
        with tempfile.NamedTemporaryFile(suffix=".jsonl") as f:
            await traj.to_jsonl(filename=f.name)
            rehydrated_traj = Trajectory.from_jsonl(f.name)
            assert traj.traj_id == rehydrated_traj.traj_id

    assert all(v == 2 for v in counter_callback.fn_invocations.values())


@pytest.mark.vcr
@pytest.mark.parametrize("fallback", [True, False])
@pytest.mark.asyncio
async def test_fallbacks_working(fallback: bool) -> None:
    AGENT_MODEL_LIST = [
        {
            "model_name": "openai/gpt-4o-mini",
            "litellm_params": {"model": "openai/gpt-4o-mini", "api_key": "abc123"},
        },
        {
            "model_name": "openai/gpt-4o",
            "litellm_params": {
                "model": "openai/gpt-4o",
            },
        },
    ]
    AGENT_ROUTER_KWARGS: dict[str, bool | list[dict[str, list[str]]]] = {
        "set_verbose": True,
    }
    if fallback:
        AGENT_ROUTER_KWARGS["fallbacks"] = [
            {
                "openai/gpt-4o-mini": [
                    "openai/gpt-4o",
                ]
            }
        ]

    AGENT_CONFIG = {
        "llm_model": {
            "name": "openai/gpt-4o-mini",
            "config": {
                "model_list": AGENT_MODEL_LIST,
                "router_kwargs": AGENT_ROUTER_KWARGS,
            },
        }
    }
    if fallback:
        AGENT_CONFIG["llm_model"]["config"]["fallbacks"] = [  # type: ignore[index]
            {
                "openai/gpt-4o-mini": [
                    "openai/gpt-4o",
                ]
            }
        ]
    agent = SimpleAgent(**AGENT_CONFIG)
    callback = DummyCallback()

    rollout_manager = RolloutManager(
        agent,
        catch_agent_failures=fallback,
        callbacks=[callback],
    )
    env = DummyEnv()
    if fallback:
        assert await rollout_manager.sample_trajectories(
            environments=[env], max_steps=2
        )
    else:
        with pytest.raises(AuthenticationError):
            await rollout_manager.sample_trajectories(environments=[env], max_steps=2)
    env.close_mock.assert_awaited_once_with()


async def adeepcopy(x):  # noqa: RUF029
    return deepcopy(x)


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_beam_search() -> None:
    agent = SimpleAgent()
    callback = DummyCallback()
    beam_search = BeamSearchRollout(
        agent,
        beam_width=1,  # keep these numbers small to speed up test
        samples_per_beam=1,
        env_clone_fn=adeepcopy,
        agent_clone_fn=deepcopy,
        scoring_fn=count_exclamations,
        catch_agent_failures=False,
        catch_env_failures=False,
        callbacks=[callback],
    )

    trajs = await beam_search.sample_trajectories(
        environments=[DummyEnv(), DummyEnv()], max_steps=1
    )
    assert len(trajs) == 2

    assert all(
        v == 2
        for k, v in callback.fn_invocations.items()
        if k != "after_agent_init_state"  # TODO: support this callback too
    )


class DummyCallback(Callback):
    def __init__(self):
        # NOTE: don't use collections.defaultdict here because it can lead to
        # test aliasing for a callback being missed altogether
        self.fn_invocations = {
            "before_rollout": 0,
            "before_transition": 0,
            "after_agent_init_state": 0,
            "after_agent_get_asv": 0,
            "after_env_reset": 0,
            "after_env_step": 0,
            "after_transition": 0,
        }

    async def before_rollout(self, traj_id: str, env: Environment) -> None:
        self.fn_invocations["before_rollout"] += 1

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        self.fn_invocations["before_transition"] += 1

    async def after_agent_init_state(self, traj_id: str, init_state: Any) -> None:
        self.fn_invocations["after_agent_init_state"] += 1

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ):
        self.fn_invocations["after_agent_get_asv"] += 1

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        self.fn_invocations["after_env_reset"] += 1

    async def after_env_step(
        self,
        traj_id: str,
        obs: list[Message],
        reward: float,
        done: bool,
        trunc: bool,
    ):
        self.fn_invocations["after_env_step"] += 1

    async def after_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        transition: Transition,
    ) -> None:
        self.fn_invocations["after_transition"] += 1


class CountingAgentState(BaseModel):
    count: float = 0.0


class CountingAgent(Agent[CountingAgentState]):
    def __init__(self):
        self.op = FxnOp[ToolRequestMessage](lambda: ToolRequestMessage(tool_calls=[]))

    async def init_state(self, tools: list[Tool]) -> CountingAgentState:
        return CountingAgentState()

    @compute_graph()
    async def get_asv(
        self, agent_state: CountingAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], CountingAgentState, float]:
        new_state = CountingAgentState(count=float(cast("str", obs[0].content)) + 1)
        action = await self.op()
        return action, new_state, 0.0


class CountingEnv(Environment[float]):
    def __init__(self, state: float = 0.0):
        self.state = state

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        return [Message(content=str(self.state))], []

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        self.state += 1
        return [Message(content=str(self.state))], 0.0, self.state >= 3, False

    def export_frame(self) -> Frame:
        return Frame()


@pytest.mark.asyncio
async def test_deterministic_rollout():
    agent = CountingAgent()
    env = CountingEnv()

    rollout_manager = RolloutManager(agent)
    traj, *_ = await rollout_manager.sample_trajectories(environments=[env])

    assert len(traj.steps) == 3
    for i_step, step in enumerate(traj.steps):
        f_step = float(i_step)
        # check that we didn't clobber any agent or env states
        assert step.agent_state.count == f_step
        assert step.next_agent_state.count == f_step + 1
        assert step.observation[0].content == str(f_step)
        assert step.next_observation[0].content == str(f_step + 1)


class NoisyCountingEnv(CountingEnv):
    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        self.state += 1 + random.uniform(-0.01, 0.01)
        return [Message(content=str(self.state))], 1.0, self.state >= 3, False


class TestTreeSearch:
    @pytest.mark.asyncio
    async def test_tree_search(self) -> None:
        agent = CountingAgent()
        # Use a slightly stochastic env so we can distinguish branches
        env = NoisyCountingEnv()

        callback = DummyCallback()
        rollout_manager = TreeSearchRollout(
            agent,
            branching_factor=2,
            env_clone_fn=deepcopy,
            concurrency_limit=1,
            callbacks=[callback],
        )
        tree = await rollout_manager.sample_tree(env, max_depth=3)
        trajs = tree.get_trajectories()
        assert len(trajs) == 8

        traj_ids_wo_root: set[str] = {
            cast("str", traj.traj_id).replace(tree.root_id, "").lstrip(":")
            for traj in trajs
        }
        # IDs should be 0:0:0, 0:0:1, ... 1:1:1 (order doesn't matter)
        assert traj_ids_wo_root == {
            ":".join(x) for x in itertools.product("01", repeat=3)
        }

        observations: dict[tuple[str, ...], str] = {}
        for traj in trajs:
            branch_path = tuple(cast("str", traj.traj_id).split(":")[1:])

            prev_step: Transition | None = None
            for i_step, step in enumerate(traj.steps):
                if prev_step is not None:
                    # Check that the child node started at the state emitted at the parent node
                    assert prev_step.next_agent_state == step.agent_state

                # Steps that started at the same node in the tree should have the same observation
                node_id = branch_path[: i_step + 1]
                assert len(step.observation) == 1
                assert step.observation[0].content
                if node_id in observations:
                    assert observations[node_id] == step.observation[0].content
                else:
                    observations[node_id] = step.observation[0].content

                prev_step = step

        for callback_fn, num_calls in callback.fn_invocations.items():
            if callback_fn in {
                "before_rollout",
                "after_agent_init_state",
                "after_env_reset",
            }:
                assert num_calls == 1, "These should be invoked once at the start"
            else:
                # We expect sum_{i=1}^3 2^i = 2^4 - 2 = 14 transitions:
                # - branching factor = 2, depth = 3
                # - root node isn't sampled, so no i=0 term in sum
                assert num_calls == 14

    @pytest.mark.asyncio
    async def test_early_stopping(self) -> None:
        agent = CountingAgent()
        # Use a slightly stochastic env so we can distinguish branches
        env = NoisyCountingEnv()

        callback = DummyCallback()
        rollout_manager = TreeSearchRollout(
            agent,
            branching_factor=2,
            env_clone_fn=deepcopy,
            concurrency_limit=1,
            callbacks=[callback],
            target_reward=0.5,
        )
        trajs = (await rollout_manager.sample_tree(env, max_depth=3)).get_trajectories()
        assert len(trajs) < 8  # should have exited early
        for traj in trajs:
            # should have hit target reward immediately
            assert len(traj.steps) == 1



import asyncio
import logging
import uuid
from collections.abc import Awaitable, Callable, Sequence
from contextlib import suppress
from typing import Any, TypeAlias

from aviary.core import Message, is_coroutine_callable
from tqdm.asyncio import tqdm

from ldp.agent import Agent
from ldp.data_structures import TransitionTree

from .callbacks import Callback
from .rollout import (
    AgentError,
    CaughtError,
    EnvError,
    RolloutManager,
    TEnv,
    reraise_exc_as,
)

logger = logging.getLogger(__name__)

TEnvCloneFn: TypeAlias = Callable[[TEnv], Awaitable[TEnv]] | Callable[[TEnv], TEnv]


class TreeSearchRollout(RolloutManager):
    def __init__(
        self,
        agent: Agent,
        branching_factor: int,
        env_clone_fn: TEnvCloneFn,
        catch_agent_failures: bool = True,
        catch_env_failures: bool = True,
        callbacks: Sequence[Callback] | None = None,
        concurrency_limit: int | None = None,
        target_reward: float | None = None,
    ):
        super().__init__(
            agent,
            catch_agent_failures=catch_agent_failures,
            catch_env_failures=catch_env_failures,
            callbacks=callbacks,
            concurrency_limit=concurrency_limit,
        )

        self.branching_factor = branching_factor
        self.target_reward = (
            target_reward if target_reward is not None else float("inf")
        )
        self.target_reward_hit: set[str] = set()

        self.env_clone_fn = env_clone_fn

    async def sample_trees(
        self,
        environments: Sequence[TEnv],
        max_depth: int | None = None,
        disable_pbar: bool = False,
    ) -> list[TransitionTree]:
        return await tqdm.gather(
            *[self.sample_tree(env, max_depth) for env in environments],
            desc="Sampling Trees",
            ncols=0,
            disable=disable_pbar,
        )

    async def sample_tree(self, env: TEnv, max_depth: int | None) -> TransitionTree:
        max_depth_f = max_depth if max_depth is not None else float("inf")
        tree = TransitionTree(root_id=str(uuid.uuid4()))

        try:
            await asyncio.gather(*[
                c.before_rollout(tree.root_id, env) for c in self.callbacks
            ])

            with reraise_exc_as(EnvError, enabled=self.catch_env_failures):
                obs, tools = await env.reset()
            await asyncio.gather(*[
                c.after_env_reset(tree.root_id, obs, tools) for c in self.callbacks
            ])

            with reraise_exc_as(AgentError, enabled=self.catch_agent_failures):
                agent_state = await self.agent.init_state(tools)
            await asyncio.gather(*[
                c.after_agent_init_state(tree.root_id, agent_state)
                for c in self.callbacks
            ])
        except CaughtError:
            return tree
        finally:
            # We need to close the env here (before descending) to avoid requiring
            # the resources of this env while making new ones in the next node.
            # Also, we double-suppress EnvError because a tree search crash here kills
            # everything and denies us passing the tree to the next step
            with (
                suppress(EnvError),
                reraise_exc_as(EnvError, enabled=self.catch_env_failures),
            ):
                await env.close()

        await self._descend(
            tree=tree,
            prev_step_id=tree.root_id,
            env=env,
            agent_state=agent_state,
            obs=obs,
            prev_timestep=-1,
            prev_cumulative_reward=0.0,
            max_depth=max_depth_f,
        )

        return tree

    async def _descend(
        self,
        tree: TransitionTree,
        prev_step_id: str,
        env: TEnv,
        agent_state: Any,
        obs: list[Message],
        prev_timestep: int,
        prev_cumulative_reward: float,
        max_depth: float,
    ) -> None:
        # Descend one level in the tree, by adding branching_factor children to the branch
        # Then, recurse on each child

        if tree.root_id in self.target_reward_hit:
            # If at least one branch hit the target reward, stop descending
            return

        timestep = prev_timestep + 1

        async def inner_descend(idx: int) -> None:
            if tree.root_id in self.target_reward_hit:
                # Check again in case the target reward was hit while waiting to start
                # this step
                return

            if is_coroutine_callable(self.env_clone_fn):
                cloned_env = await self.env_clone_fn(env)
            else:
                cloned_env = self.env_clone_fn(env)

            # Descend one step
            step_id = f"{prev_step_id}:{idx}"
            try:
                step = await self._take_step(
                    timestep, step_id, cloned_env, agent_state, obs
                )
            except CaughtError:
                # If we failed, do not extend the branch - just give up on this path
                return

            if timestep + 1 == max_depth and not step.done:
                # Mark as truncated if we hit max_steps and the state is not terminal.
                step.truncated = True

            await asyncio.gather(*[
                callback.after_transition(step_id, self.agent, cloned_env, step)
                for callback in self.callbacks
            ])

            tree.add_transition(step_id, step)

            cumulative_reward = prev_cumulative_reward + step.reward
            if cumulative_reward >= self.target_reward:
                # signal other descents to stop too
                self.target_reward_hit.add(tree.root_id)
                return

            if step.done or step.truncated:
                return

            # Recurse
            await self._descend(
                tree=tree,
                prev_step_id=step_id,
                env=cloned_env,
                agent_state=step.next_agent_state,
                obs=step.next_observation,
                prev_timestep=timestep,
                prev_cumulative_reward=cumulative_reward,
                max_depth=max_depth,
            )

        # Add branching_factor children
        await asyncio.gather(*[
            inner_descend(idx) for idx in range(self.branching_factor)
        ])



from collections.abc import Iterable, Sequence
from typing import cast
from uuid import UUID

import litellm
import pytest
import tenacity
import tree
from aviary.core import Message
from lmi import CommonLLMNames
from lmi import LiteLLMModel as LLMModel
from pydantic import BaseModel, Field, JsonValue

from ldp.agent import Agent, MemoryAgent, ReActAgent
from ldp.alg.optimizer import (
    MemoryFactory,
    MemoryOpt,
    Optimizer,
    default_optimizer_factory,
)
from ldp.alg.optimizer.ape import APEOpt, APEScoreFn, Example
from ldp.data_structures import Trajectory, Transition
from ldp.graph import (
    CallID,
    FxnOp,
    LLMCallOp,
    Memory,
    MemoryOp,
    Op,
    OpCtx,
    OpResult,
    PromptOp,
    compute_graph,
    eval_mode,
)
from ldp.graph.gradient_estimators import (
    llm_straight_through_estimator as llm_ste,
)
from ldp.graph.gradient_estimators import (
    straight_through_estimator as ste,
)
from ldp.graph.ops import GradInType
from ldp.llms.prompts import append_to_sys
from tests.conftest import VCR_DEFAULT_MATCH_ON


@pytest.mark.parametrize(
    ("agent_cls", "optimizer_cls", "optimizer_kwargs"),
    [
        (MemoryAgent, MemoryOpt, {}),
        (ReActAgent, APEOpt, {"score_fn": APEScoreFn.GRADIENT}),
    ],
)
def test_optimizer_factory(
    agent_cls: type[Agent], optimizer_cls: type[Optimizer], optimizer_kwargs: dict
):
    agent = agent_cls()
    opt = default_optimizer_factory(agent, optimizer_cls, **optimizer_kwargs)
    assert isinstance(opt, optimizer_cls)


class SquaredErrorLoss(Op[int]):
    async def forward(self, y: str, yhat: str) -> int:
        try:
            return (int(y) - int(yhat)) ** 2
        except ValueError:  # For example, yhat may be "I guess the number is 7."
            return 100

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        try:
            y = int(input_kwargs["y"])
            yhat = int(input_kwargs["yhat"])
        except ValueError:
            loss = ctx.get(call_id, "output").value
            return [], {"y": loss, "yhat": loss}  # Straight-through approximation
        # d/dy of (y - y^)^2 = 2 (y - y^), and d/dy^ of (y - y^)^2 = -2 (y - y^)
        # return  dL/dy,  dL/dy^
        # Note that grad_output is ignored because this is assumed to be a terminal scalar,
        # much like calling loss.backward() in pytorch.
        return [], {
            "y": 2 * (y - yhat),
            "yhat": -2 * (y - yhat),
        }


@pytest.mark.asyncio
async def test_ape_optimizer() -> None:
    sys_prompt_op = PromptOp("Guess a number based on the input word.")
    package_msg_op = FxnOp(append_to_sys)
    config = {"max_retries": 3}  # we seem to be hitting rate limits frequently
    llm = LLMModel(config=config)
    llm_call_op = LLMCallOp()
    strip_op = FxnOp(lambda x: x.content)
    loss_op = SquaredErrorLoss()

    @compute_graph()
    async def forward(xi_: str, yi_: str) -> OpResult[int]:
        """Perform a forward pass through the model to the resultant SE loss."""
        s = await sys_prompt_op()
        m = await package_msg_op(xi_, s)
        c = await llm_call_op(config, m)
        yh = await strip_op(c)
        return await loss_op(yi_, yh)

    # Sequentially run a forward pass for each (x, y)
    x = ["Hello", "Day", "Bar"]
    y = [str(len(xi)) for xi in x]  # Number to guess should be word's length
    opt = APEOpt(
        llm=llm,
        llm_call_op=llm_call_op,
        prompt_op=sys_prompt_op,
        good_examples=[
            Example(input=x, output=y, score=0) for x, y in zip(x, y, strict=True)
        ],
        score_fn=APEScoreFn.GRADIENT,
    )
    assert opt.trace == [sys_prompt_op.prompt]

    trajectory = Trajectory()
    for i, (xi, yi) in enumerate(zip(x, y, strict=True)):
        loss = await forward(xi, yi)
        if i == 0:
            assert loss.value > 0, (
                "First example's loss should be non-zero - otherwise, no learning"
                " signal."
            )
        # Sets grad_output and grad_input in context, to be used by optimizer
        loss.compute_grads(backward_fns={LLMCallOp: llm_ste, FxnOp: ste})

        # APE in gradient mode is only going to pay attention to the action, so set
        # placeholders for the other attributes
        trajectory.steps.append(
            Transition(
                timestep=0,
                agent_state=None,
                next_agent_state=None,
                observation=[],
                next_observation=Transition.NO_OBSERVATION,
                action=loss,
                reward=0,
                done=False,
            )
        )

    # Run full optimizer step
    for i in range(3):  # Retries
        opt.aggregate([trajectory])
        assert opt.good_examples == [
            Example(input=x, output=y, score=0) for x, y in zip(x, y, strict=True)
        ]

        await opt.update()
        assert not opt.examples, "Expected reset of examples after update."
        assert len(opt.trace) == i + 2, "Expected new prompt to be recorded."

        with eval_mode():
            if (await forward(xi, yi)).value == 0:  # pylint: disable=undefined-loop-variable
                return

    raise AssertionError("Failed to complete optimization after retries.")


class NumberGuesserModule:
    """Made up module used to enable simple training scripts."""

    def __init__(self):
        self.mem_op = MemoryOp()
        self.package_msg_op = FxnOp(self._package)
        self.llm_call_op = LLMCallOp()
        self.strip_op = FxnOp(lambda x: x.content)

    @staticmethod
    def _package(mems: Iterable[Memory], query: str) -> list[Message]:
        itemized_mems = "\n\n".join(str(m) for m in mems)
        return [
            Message(content="Guess a number based on the input word."),
            Message(content=f"Previous attempts:\n{itemized_mems}\n-----\n\n{query}"),
        ]

    async def __call__(self, query: str) -> tuple[OpResult[str], list[Message]]:
        mems = await self.mem_op(query)
        msgs = await self.package_msg_op(mems, query)
        c = await self.llm_call_op(
            config={
                "name": "gpt-4-turbo",  # this is flaky, so use a smarter model
                "temperature": 0,
                "max_retries": 3,
            },
            msgs=msgs,
        )
        return await self.strip_op(c), msgs.value


async def nondifferentiable_reward_model(target: str, result: OpResult[str]) -> float:
    se = (await SquaredErrorLoss()(target, result)).value
    if se == 0:
        return 1.0  # Positive reward if it got it right
    return -se  # Squared error is positive, so convert to negative


class TestMemoryOpt:
    @staticmethod
    def _mem_opt_failed(exc: BaseException) -> bool:
        # Sometimes the memory opt fails to converge because the training examples
        # are not informative. Try again
        return isinstance(exc, AssertionError) and "should be less than" in str(exc)

    @pytest.mark.flaky(reruns=3, only_on=[litellm.exceptions.APIConnectionError])
    @pytest.mark.asyncio
    @tenacity.retry(
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception(_mem_opt_failed),
    )
    async def test_standard_memory_optimizer(self) -> None:
        model = NumberGuesserModule()
        # seed with one memory to show example
        await model.mem_op.memory_model.add_memory(
            Memory(
                query="Great",
                output=str(len("Great")),
                value=1.0,
                metadata={"timestep": 0, "done": False, "truncated": False},
            )
        )
        prior_num_memories = 1
        opt = MemoryOpt(memory_op=model.mem_op, output_op=model.llm_call_op)

        x = ["Hello", "Day", "Bar"]
        y = [str(len(xi)) for xi in x]
        trajectory = Trajectory()
        for xi, yi in zip(x, y, strict=True):
            async with compute_graph():
                yh, _ = await model(xi)
                # MemoryOp works with rewards, not gradients. So instead of backpropagating
                # through the loss, for training we compute a non-differentiable reward.
                reward = await nondifferentiable_reward_model(yi, yh)
            yh.compute_grads()

            # MemoryOpt is only going to look at action and reward,
            # so set placeholders for the other values
            trajectory.steps.append(
                Transition(
                    timestep=0,
                    agent_state=None,
                    next_agent_state=None,
                    observation=Transition.NO_OBSERVATION,
                    next_observation=Transition.NO_OBSERVATION,
                    action=yh,
                    reward=reward,
                    done=False,
                )
            )

        opt.aggregate([trajectory])
        await opt.update()

        assert len(model.mem_op.memory_model.memories) == len(x) + prior_num_memories, (
            "Incorrect number of stored memories after optimization step."
        )
        assert all(
            not cast("dict", m.metadata)["done"]
            for m in model.mem_op.memory_model.memories.values()
        )
        assert not opt.example_buffer, (
            "MemoryOpt buffer should be empty after applying update"
        )

        x_eval, y_eval = xi, yi  # pylint: disable=undefined-loop-variable
        async with compute_graph():
            with eval_mode():
                yh, msgs = await model(x_eval)
            assert len(msgs) > 1, "Message should have multiple memories."
            # check that Input appears a few times (from memories)
            assert msgs[-1].content, "unexpected message content"
            assert msgs[-1].content.count("Input") > 2, (
                "Input should appear multiple times in the response."
            )
            se_loss = (await SquaredErrorLoss()(y_eval, yh)).value

        assert se_loss < 100, (
            f"Loss ({se_loss:.3f}) should be less than 100 after training."
        )

    @pytest.mark.vcr(match_on=[*VCR_DEFAULT_MATCH_ON, "body"])
    @pytest.mark.asyncio
    async def test_lessons_memory_optimizer(self) -> None:
        """
        Test we can use LLM completions to generate lessons instead of memories.

        This test is loosely based on Reflexion (https://arxiv.org/abs/2303.11366).
        """
        memory_distiller = LLMModel(config={"name": CommonLLMNames.OPENAI_TEST.value})

        class LessonEntry(BaseModel):
            """Entry for a lesson created from some example data."""

            query: str = Field(
                description=(
                    "Plain text string for retrieving this lesson from a database of"
                    " lessons."
                )
            )
            lesson: str = Field(description="Lesson generated from past attempts.")

            @staticmethod
            def make_prompt(examples: Sequence[tuple]) -> str:
                """Create an LLM prompt to generate a lesson from examples."""
                itemized_examples = "\n-".join(str(x) for x in examples)
                return (
                    "We are trying to guess a number based on the input word. We just"
                    f" tried {len(examples)} times, and collected rewards where a"
                    " higher reward is better. Here are the results in three-tuples of"
                    f" input, output, reward:\n- {itemized_examples}\n\nPlease create"
                    " a lesson based on this data referencing the relative success or"
                    " failure associated with the reward, use concise wording and"
                    " don't repeat yourself."
                )

            @classmethod
            def to_memory(cls, lesson_json: str, run_id: UUID | None = None) -> Memory:
                lesson = cls.model_validate_json(lesson_json)
                return Memory(
                    query=lesson.query, output=lesson.lesson, value="", run_id=run_id
                )

            @classmethod
            async def memory_factory(
                cls,
                memory_op: MemoryOp,
                output_op: Op[Message],
                memory_template: str,
                example_buffer: Iterable[tuple[CallID, CallID, float, JsonValue]],
            ) -> list[Memory]:
                example_buffer = list(example_buffer)
                if not example_buffer:
                    raise RuntimeError("Expected non-empty example buffer.")
                query_airesponse_dreturns: list[tuple[str, str, float]] = [
                    (
                        memory_op.ctx.get(mem_call_id, "query"),
                        output_op.ctx.get(output_call_id, "output").value.content,
                        d_return,
                    )
                    for mem_call_id, output_call_id, d_return, metadata in example_buffer
                ]
                response = await memory_distiller.call_single(
                    messages=[
                        Message(
                            content=LessonEntry.make_prompt(query_airesponse_dreturns)
                        )
                    ],
                    tool_choice=memory_distiller.NO_TOOL_CHOICE,
                    output_type=LessonEntry,
                )
                if (
                    not response.messages
                    or len(response.messages) != 1
                    or not response.messages[0].content
                ):
                    raise ValueError(
                        "Expected a single message in the response containing a"
                        f" serialized {cls.__name__}."
                    )
                return [
                    cls.to_memory(
                        response.messages[0].content, run_id=example_buffer[0][0].run_id
                    )
                ]

        assert isinstance(LessonEntry.memory_factory, MemoryFactory)

        model = NumberGuesserModule()
        opt = MemoryOpt(
            memory_op=model.mem_op,
            output_op=model.llm_call_op,
            memory_factory=LessonEntry.memory_factory,
        )

        x = ["Hello", "Day", "Bar"]
        y = [str(len(xi)) for xi in x]
        trajectory = Trajectory()
        for xi, yi in zip(x, y, strict=True):
            async with compute_graph():
                yh, _ = await model(xi)
                # MemoryOp works with rewards, not gradients. So instead of backpropagating
                # through the loss, for training we compute a non-differentiable reward.
                reward = await nondifferentiable_reward_model(yi, yh)
            yh.compute_grads()

            # MemoryOpt is only going to look at action and reward,
            # so set placeholders for the other values
            trajectory.steps.append(
                Transition(
                    timestep=0,
                    agent_state=None,
                    next_agent_state=None,
                    observation=Transition.NO_OBSERVATION,
                    next_observation=Transition.NO_OBSERVATION,
                    action=yh,
                    reward=reward,
                    done=False,
                )
            )

        opt.aggregate([trajectory])
        await opt.update()

        assert not opt.example_buffer, (
            "MemoryOpt buffer should be empty after applying update"
        )



from typing import TypeVar

import torch
from transformers.generation.utils import GenerateDecoderOnlyOutput

TOutputType = TypeVar("TOutputType", torch.Tensor, GenerateDecoderOnlyOutput)


class TensorChunker:
    """Splits tensors into chunks and adds dummy chunks as needed for parallel processing frameworks like FSDP."""

    def __init__(self, num_chunks: int):
        self.num_chunks = num_chunks

    def chunkify(self, *args, **kwargs) -> tuple[list[tuple], list[dict], list[bool]]:
        """Splits the args into self.num_chunks chunks, adding dummy chunks as needed.

        Returns:
            tuple: Contains:
                - A list of len self.num_chunks, where each element is a tuple of split args.
                - A list of len self.num_chunks, where each element is a dict of split kwargs.
                - A list of len self.num_chunks, where each element is a bool indicating
                  whether the chunk is a dummy chunk.

        NOTE: the below Note is not always true - should come back and fix.
            Can be seen by observing e.g. `len(torch.chunk(torch.arange(20), 8, dim=0)) == 7`.
            Not necessarily a huge problem - just means we'll be chunking suboptimally.
            But there may be cases where that causes problems.
        # Note: Dummy flags will be set only if the incoming batch is smaller than self.num_chunks,
        # resulting in empty chunks insertion.

        Example:
        ```python
        input_tensor = torch.randn(10, 5)  # A tensor with batch size 10
        some_int = 42  # An integer value that isn't split

        chunker = TensorChunker(num_chunks=4)
        split_args, split_kwargs, dummy_flags = chunker.chunkify(
            input_tensor=input_tensor, some_int=some_int
        )
        ```

        Output split_args:
        [
            ((input_tensor_chunk_0, ), (some_int, )),  # batch of 3
            ((input_tensor_chunk_1, ), (some_int, )),  # batch of 3
            ((input_tensor_chunk_2, ), (some_int, )),  # batch of 3
            ((input_tensor_chunk_3, ), (some_int, )),  # batch of 1
        ],
        dummy_flags = [False, False, False, False]

        """
        # Initialize lists to hold split args and kwargs
        split_args_list: list[list] = [[] for _ in range(self.num_chunks)]
        split_kwargs_list: list[dict] = [{} for _ in range(self.num_chunks)]
        dummy_chunk_flags: list[bool] = []

        def process_items(items, is_kwargs=False):
            """Helper function to process both args and kwargs."""
            nonlocal dummy_chunk_flags
            for item in items:
                key, value = item if is_kwargs else (None, item)
                chunks, out_chunk_dummy_flags = self._split_value(value)
                for i in range(self.num_chunks):
                    if is_kwargs:
                        split_kwargs_list[i][key] = chunks[i]
                    else:
                        split_args_list[i].append(chunks[i])

                # Ensure dummy chunk flags are consistent across all items
                if dummy_chunk_flags and out_chunk_dummy_flags:
                    assert dummy_chunk_flags == out_chunk_dummy_flags, (
                        "All inputs must result in the same dummy chunk flags."
                    )
                dummy_chunk_flags = out_chunk_dummy_flags or dummy_chunk_flags

        # Process args and kwargs
        process_items(args)
        process_items(kwargs.items(), is_kwargs=True)

        split_args_tuples = [tuple(args) for args in split_args_list]
        return split_args_tuples, split_kwargs_list, dummy_chunk_flags

    @staticmethod
    def dechunkify(
        outputs_list: list[TOutputType],
        dummy_chunk_flags: list[bool],
    ) -> TOutputType:
        """Reassembles the outputs from the handlers, removing outputs corresponding to dummy chunks.

        The desired behavior is such that the following are equivalent:
        ```python
        # "single device"
        output_single = module(input)

        # "multi device" (in reality each device would only run one chunk)
        chunks, dummy_chunk_flags = chunker.chunkify(input)
        output_multi = chunker.dechunkify(
            [module(chunk) for chunk in chunks], dummy_chunk_flags
        )
        ```
        """
        # Filter out outputs corresponding to dummy chunks
        real_outputs = [
            output
            for output, is_dummy in zip(outputs_list, dummy_chunk_flags, strict=True)
            if not is_dummy
        ]

        if isinstance(real_outputs[0], torch.Tensor):
            # If we received a list of tensors, concat along the batch dimension
            # TODO: consider how/if to handle differently shaped tensors
            return torch.cat(real_outputs, dim=0)

        if isinstance(real_outputs[0], GenerateDecoderOnlyOutput):
            sequences: list[torch.Tensor] = []
            for output in real_outputs:
                sequences.extend(output.sequences)
            scores: list[list[torch.Tensor | None]] | None = None
            if real_outputs[0].scores is not None:
                assert all(output.scores is not None for output in real_outputs)
                # `output.scores` is a tuple (one el per decoded token), with each tensor of shape
                # (batch_size, vocab_size).
                # To get the scores for the `i`-th batch element, we'd do `[score[i] for score in output.scores]`.
                # `output.scores` will have different lengths across workers, so we cannot simply concatenate them.
                # Adding dummy scores is error-prone and loses info. Instead, we create a merged `scores` with the
                # same access semantics. `score[i]` will be a tensor if it exists for batch element `i` and `None`
                # otherwise.
                scores = []
                max_output_len = max(
                    len(output.scores or []) for output in real_outputs
                )
                for i in range(max_output_len):
                    scores_step_i: list[torch.Tensor | None] = []
                    for output in real_outputs:
                        if output.scores is not None and i < len(output.scores):
                            scores_step_i.extend(list(output.scores[i]))
                        else:
                            # Add bsz Nones for this worker, since it did not reach token `i`.
                            scores_step_i.extend([None] * output.sequences.shape[0])

                    scores.append(scores_step_i)

            return GenerateDecoderOnlyOutput(
                sequences=sequences,  # type: ignore[arg-type]
                scores=scores,  # type: ignore[arg-type]
            )

        raise ValueError("Unsupported output type")

    def _split_value(self, value):
        """
        Splits a single value into chunks, adding a dummy value if necessary.

        Right now, only torch.Tensor values are split. Non-tensor values are replicated.
        """
        if isinstance(value, torch.Tensor):
            chunks = list(torch.chunk(value, self.num_chunks, dim=0))
            dummy_chunk_flags = []
            for i in range(self.num_chunks):
                if i >= len(chunks):
                    # Chunk 0 will always exist, and we need only a batch of one ([:1])
                    # to activate the model.
                    # We use the first element of the existing chunks as real data to avoid
                    # errors in the model that may expect a specific token structure.
                    chunks.append(chunks[0][:1])
                    dummy_chunk_flags.append(True)
                else:
                    dummy_chunk_flags.append(False)

            return chunks, dummy_chunk_flags
        # Non-tensor values are replicated
        return [value] * self.num_chunks, None



import operator

import numpy as np
import pytest
from aviary.core import DummyEnv
from aviary.utils import MultipleChoiceQuestion

from ldp.agent import SimpleAgent
from ldp.alg import bulk_evaluate_consensus, compute_pass_at_k
from ldp.alg.algorithms import measure_consensus_proportion
from ldp.utils import discounted_returns


@pytest.mark.asyncio
async def test_rollout_and_discounting(dummy_env: DummyEnv) -> None:
    obs, tools = await dummy_env.reset()

    agent = SimpleAgent(tools=tools)
    agent_state = await agent.init_state(tools=tools)

    observations = []
    actions = []
    rewards = []
    terms = []
    done = True
    for i in range(3):  # noqa: B007
        if done:
            obs, _ = await dummy_env.reset()
            agent_state = await agent.init_state(tools=tools)

        observations.append((obs, agent_state))
        action, agent_state, _ = await agent.get_asv(agent_state, obs)
        obs, reward, done, _ = await dummy_env.step(action.value)
        actions.append(action)
        rewards.append(reward)
        terms.append(done)

    print(terms)
    d_returns = discounted_returns(rewards, terms, 0.5)
    print(d_returns)


@pytest.mark.asyncio
async def test_consensus_evaluation() -> None:
    # We have two questions, so let's group based on question
    question_1 = MultipleChoiceQuestion(
        question="What is the meaning of life?",
        options=["-84", "11", "cheesecake"],
        ideal_answer="42",
    )
    question_2 = MultipleChoiceQuestion(
        question="What is a healthy fruit?",
        options=["brownie", "chocolate bar", "french fry"],
        ideal_answer="apple",
    )
    question_3 = MultipleChoiceQuestion(
        question="What is the highest number?",
        options=["1", "2", "4"],
        ideal_answer="8",
    )
    data_with_several_groups: list[tuple[MultipleChoiceQuestion, str]] = [
        # Has consensus and it was correct
        (question_1, "-84"),
        (question_1, "11"),
        (question_1, "11"),
        (question_1, "cheesecake"),
        (question_1, "42"),
        (question_1, "42"),
        (question_1, "42"),
        (question_1, "42"),
        (question_1, "42"),
        (question_1, "42"),
        # Has consensus and it was correct
        (question_2, "brownie"),
        (question_2, "brownie"),
        (question_2, "apple"),
        (question_2, "apple"),
        (question_2, "apple"),
        (question_2, "apple"),
        (question_2, "apple"),
        (question_2, "apple"),
        # Has no consensus and regardless it's incorrect
        (question_3, "1"),
        (question_3, "2"),
        (question_3, "1"),
        (question_3, "2"),
        (question_3, "4"),
    ]
    # NOTE: this consensus is sensitive to seed
    expected_consensus = {
        question_1.question: (
            [("42", 3), ("cheesecake", 1), ("-84", 1)],
            3 / 5,
            0.2190890,
        ),
        question_2.question: ([("apple", 4), ("brownie", 1)], 4 / 5, 0.1788854),
        question_3.question: ([("2", 2), ("1", 2), ("4", 1)], 2 / 5, 0.2190890),
    }
    stored_accuracy_mean_ste: list[tuple[float, float]] = []

    def append_accuracy_metrics(
        consensus_answer: str,  # noqa: ARG001
        consensus_size: int,
        sample_size: int,
    ) -> None:
        stored_accuracy_mean_ste.append(
            measure_consensus_proportion(consensus_size, sample_size)
        )

    # Check accuracy is 0% without an ideal answer
    groups, accuracy = await bulk_evaluate_consensus(
        data_with_several_groups,
        grouping_fn=lambda x: x[0].question,
        num_samples=5,
        seed=np.random.default_rng(42),
        extract_answer_fn=operator.itemgetter(1),
        consensus_callback=append_accuracy_metrics,
    )
    assert len(groups) == len(expected_consensus)
    for (q, (consensus, acc_mean, acc_ste)), actual_acc_ste in zip(
        expected_consensus.items(), stored_accuracy_mean_ste, strict=True
    ):
        assert groups[q] == consensus
        assert actual_acc_ste == (pytest.approx(acc_mean), pytest.approx(acc_ste))
    assert accuracy == 0.0, "Can't compute accuracy without an ideal answer"
    stored_accuracy_mean_ste.clear()  # Prepare for next batch of assertions

    # Check accuracy is present when we can get an ideal answer
    groups, accuracy = await bulk_evaluate_consensus(
        data_with_several_groups,
        grouping_fn=lambda x: x[0].question,
        ideal_answer_fn=lambda x: x[0].ideal_answer,
        num_samples=5,
        seed=np.random.default_rng(42),
        extract_answer_fn=operator.itemgetter(1),
        consensus_callback=append_accuracy_metrics,
    )
    assert len(groups) == len(expected_consensus)
    for (q, (consensus, acc_mean, acc_ste)), actual_acc_ste in zip(
        expected_consensus.items(), stored_accuracy_mean_ste, strict=True
    ):
        assert groups[q] == consensus
        assert actual_acc_ste == (pytest.approx(acc_mean), pytest.approx(acc_ste))
    assert accuracy == 2 / 3

    with pytest.raises(ValueError, match="sampling with replacement"):
        await bulk_evaluate_consensus(
            data_with_several_groups,
            grouping_fn=lambda x: x[0].question,
            num_samples=10,  # Sampling with replacement is disallowed
            extract_answer_fn=operator.itemgetter(1),
        )


@pytest.mark.parametrize(
    ("n", "c", "k", "expected"),
    [
        pytest.param(10, 10, 3, 1.0, id="all-correct-k>1"),
        pytest.param(10, 10, 1, 1.0, id="all-correct-k=1"),
        pytest.param(10, 0, 3, 0.0, id="all-incorrect-k>1"),
        pytest.param(10, 0, 1, 0.0, id="all-incorrect-k=1"),
        pytest.param(3, 1, 3, 1.0, id="n-c<k"),
        (10, 5, 3, 1 - 1 / 12),
        (2, 1, 1, 1 / 2),  # Match https://ai.stackexchange.com/a/40396
        # SEE: https://github.com/parker-research/pass-at-k/blob/037c5d477486f9e95e1c21fc349576447cd6ce8b/tests/test_pass_at_k.py
    ],
)
def test_compute_pass_at_k(n: int, c: int, k: int, expected: float) -> None:
    assert compute_pass_at_k(n, c, k) == pytest.approx(expected)



import numpy as np
import pytest
import torch

from ldp.graph import MSELossOp, compute_graph


@pytest.mark.asyncio
@pytest.mark.parametrize("input_size", [4, 10])
@pytest.mark.parametrize("dtype", ["numpy", "torch"])
async def test_embedding_op(input_size, dtype) -> None:
    op = MSELossOp()

    # Generate data based on dtype
    if dtype == "numpy":
        rng = np.random.default_rng(12345)
        prediction = rng.random(input_size)
        target = rng.random(input_size)
    else:
        prediction = torch.rand(input_size)
        target = torch.rand(input_size)
    async with compute_graph():
        op_result = await op(
            prediction=prediction,
            target=target,
        )

    # Validate the output and grads
    op_result.compute_grads()
    grads = op.get_input_grads(op_result.call_id)
    assert grads[0] == []
    assert grads[1].keys() == {"prediction", "target"}
    assert grads[1].get("target") is None
    pred = grads[1].get("prediction")
    if dtype == "numpy":
        assert isinstance(pred, np.ndarray)
        assert isinstance(op_result.value, float)
    else:
        assert isinstance(pred, torch.Tensor)
        assert isinstance(op_result.value, torch.Tensor)
    assert pred.shape == (input_size,)



from abc import ABC, abstractmethod

import torch
from transformers import LogitsProcessor


class LogitsProcessorWithFinalize(LogitsProcessor, ABC):
    @abstractmethod
    def finalize(self, token_ids: torch.Tensor) -> None:
        """A method for subclasses to inject arbitrary finalization logic after sampling finishes.

        TransformerHandler will invoke logit_processor.finalize(token_ids), where token_ids are
        the sampled tokens.
        """



import logging
from typing import Any, Self, cast

from aviary.core import (
    MalformedMessageError,
    Message,
    Tool,
    ToolRequestMessage,
    ToolResponseMessage,
)
from lmi import CommonLLMNames
from pydantic import BaseModel, ConfigDict, Field
from tenacity import (
    Future,
    RetryCallState,
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
)

from ldp.graph import OpResult, compute_graph
from ldp.graph.modules.react import (
    ACT_DEFAULT_PROMPT_TEMPLATE,
    ACT_DEFAULT_SINGLE_PROMPT_TEMPLATE,
    REACT_DEFAULT_PROMPT_TEMPLATE,
    REACT_DEFAULT_SINGLE_PROMPT_TEMPLATE,
    ReActModule,
    ReActModuleSinglePrompt,
    ToolDescriptionMethods,
)

from . import DEFAULT_LLM_COMPLETION_TIMEOUT
from .agent import Agent
from .simple_agent import SimpleAgentState

logger = logging.getLogger(__name__)


class ReActAgent(BaseModel, Agent[SimpleAgentState]):
    """An Act or ReAct Agent built to work with chat models.

    Paper: https://arxiv.org/abs/2210.03629

    The ReAct style is like so, and note Act style has no 'Thought: ' entries:
    System:
        Answer the following questions as best you can. You have access to the following tools:

            {tools}

            Use the following format:

            Thought: you should always think about what to do
            Action: the action to take, should be one of [{tool_names}]
            Action Input: the input to the action
            Observation: the result of the action
            ... (this Thought/Action/Action Input/Observation can repeat N times)
    User:
        {questions}
    Assistant:
        Thought:
        Action:
        Action Input:
    User:
        Observation:
    Assistant:
        Thought:
        Action:
        Action Input:
    ...

    One notable design decision is that ReAct's state does not necessarily
    track ToolRequestMessage. Recall that aviary is in a partially observable
    domain, meaning we don't need to have perfect symmetry with Environments.
    Instead, ReActAgent's state stores a ReAct-style message history, where the
    messages are plain Message (and not a ToolRequestMessage).
    """

    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    llm_model: dict[str, Any] = Field(
        default={
            "name": CommonLLMNames.GPT_4O.value,
            "temperature": 0.1,
            "logprobs": True,
            "top_logprobs": 1,
            "timeout": DEFAULT_LLM_COMPLETION_TIMEOUT,
        },
        description="Starting configuration for the LLM model.",
    )
    sys_prompt: str = Field(
        description=(
            "Learnable system prompt template. If not provided, a default ReAct prompt "
            "template will be assigned, depending on the single_prompt setting."
        ),
    )
    tool_description_method: ToolDescriptionMethods = Field(
        default=ToolDescriptionMethods.STR,
        description="Method used to describe the tools, defaults to 'str' description.",
    )
    single_prompt: bool = Field(
        default=False,
        description=(
            "Specifies whether to use a single prompt for both reasoning and action"
            " selection, or to use 2 sequential prompts. When set to True, a single API"
            " call is made, and ldp handles the message parsing to extract the action."
            " If set to False, a second API call is made specifically to request the"
            " action, with parsing done on the API side. Defaults to False, as it"
            " results in fewer action selection failures."
        ),
    )

    hide_old_env_states: bool = Field(
        default=False,
        description="See SimpleAgentState.hide_old_env_states.",
    )

    hide_old_action_content: bool = Field(
        default=False,
        description="See SimpleAgentState.hide_old_action_content.",
    )

    sliding_window: int | None = Field(
        default=None,
        description="See SimpleAgentState.sliding_window.",
    )

    @classmethod
    def make_act_agent(cls, **kwargs) -> Self:
        single_prompt = kwargs.pop("single_prompt", False)
        return cls(
            sys_prompt=(
                ACT_DEFAULT_SINGLE_PROMPT_TEMPLATE
                if single_prompt
                else ACT_DEFAULT_PROMPT_TEMPLATE
            ),
            **kwargs,
        )

    def __init__(self, **kwargs):
        # set sys_prompt if not provided
        if "sys_prompt" not in kwargs:
            single_prompt = kwargs.get("single_prompt", False)
            kwargs["sys_prompt"] = (
                REACT_DEFAULT_SINGLE_PROMPT_TEMPLATE
                if single_prompt
                else REACT_DEFAULT_PROMPT_TEMPLATE
            )

        super().__init__(**kwargs)
        if self.single_prompt:
            self._react_module = ReActModuleSinglePrompt(
                self.llm_model, self.sys_prompt, self.tool_description_method
            )
        else:
            self._react_module = ReActModule(
                self.llm_model, self.sys_prompt, self.tool_description_method
            )

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        return SimpleAgentState(
            tools=tools,
            hide_old_env_states=self.hide_old_env_states,
            hide_old_action_content=self.hide_old_action_content,
            sliding_window=self.sliding_window,
        )

    @staticmethod
    def after_retry_failure_log(retry_state: RetryCallState):
        logger.error(
            f"Failed across {retry_state.attempt_number} attempts to run get_asv given"
            f" arguments {retry_state.args} and kwargs {retry_state.kwargs}."
        )
        # NOTE: this blows up with the underlying exception... it isn't wrapped in a
        # RetryError like normal tenacity
        return cast("Future", retry_state.outcome).result()

    @retry(
        retry=retry_if_exception_type(MalformedMessageError),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        stop=stop_after_attempt(5),
        retry_error_callback=after_retry_failure_log,
    )
    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        obs = obs.copy()  # Keep original obs, as we edit the content below
        if self.single_prompt:
            for i, m in enumerate(obs):
                if isinstance(m, ToolResponseMessage):
                    obs[i] = Message(content=f"Observation: {m.content}")
        else:
            for i, m in enumerate(obs):
                if isinstance(m, ToolResponseMessage):
                    obs[i] = m.model_copy(
                        update={"content": f"Observation: {m.content}"}
                    )
        next_state = agent_state.get_next_state(obs=obs)
        action_selection_result, new_messages = await self._react_module(
            messages=next_state.messages, tools=next_state.tools
        )
        next_state.messages = [*next_state.messages, *new_messages]
        return action_selection_result, next_state, 0.0



import random
from pathlib import Path

import numpy as np
import torch


def set_seed(seed: int | None) -> None:
    if seed is None:
        return

    random.seed(seed)
    np.random.seed(seed)  # noqa: NPY002
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


REPO_ROOT = Path(__file__).parent.parent.parent



import asyncio
from functools import partial
from itertools import starmap
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp

import pytest
import torch
from torch.distributed.fsdp import StateDictType
from torch.optim import SGD
from transformers.generation.utils import GenerateDecoderOnlyOutput

import ldp.nn
from ldp.nn.handlers.transformer_handler import get_unused_port

from .conftest import PARALLEL_MODE_CONFIGS, TEST_GPUS


def test_model_load():
    config = ldp.nn.LMConfig(model="gpt2")
    config.resolve_model_location()
    config.get_causal_lm()


class TestTensorChunker:
    def test_chunkify_add_dummy_chunks(self):
        batch_size = 3
        num_chunks = 5

        sample_tensor = torch.arange(1, batch_size * 10 + 1).reshape(batch_size, 10)

        chunker = ldp.nn.TensorChunker(num_chunks=num_chunks)
        split_args, split_kwargs, dummy_chunk_flags = chunker.chunkify(sample_tensor)

        assert len(split_args) == num_chunks
        assert len(split_kwargs) == num_chunks
        assert dummy_chunk_flags == [False, False, False, True, True]
        assert torch.equal(split_args[0][0], sample_tensor[:1])
        assert torch.equal(split_args[1][0], sample_tensor[1:2])
        assert torch.equal(split_args[2][0], sample_tensor[2:3])
        assert torch.equal(split_args[3][0], sample_tensor[:1])
        assert torch.equal(split_args[4][0], sample_tensor[:1])

    def test_chunkify_no_dummy_chunks(self):
        batch_size = 9
        num_chunks = 5

        sample_tensor = torch.arange(1, batch_size * 10 + 1).reshape(batch_size, 10)

        chunker = ldp.nn.TensorChunker(num_chunks=num_chunks)
        split_args, split_kwargs, dummy_chunk_flags = chunker.chunkify(sample_tensor)

        assert len(split_args) == num_chunks
        assert len(split_kwargs) == num_chunks
        assert dummy_chunk_flags == [False, False, False, False, False]
        assert torch.equal(split_args[0][0], sample_tensor[:2])
        assert torch.equal(split_args[1][0], sample_tensor[2:4])
        assert torch.equal(split_args[2][0], sample_tensor[4:6])
        assert torch.equal(split_args[3][0], sample_tensor[6:8])
        assert torch.equal(split_args[4][0], sample_tensor[8:])

    def test_chunkify_with_args_and_kwargs(self):
        batch_size = 2
        num_chunks = 3

        sample_tensor = torch.arange(1, batch_size * 10 + 1).reshape(batch_size, 10)
        sample_tensor_kwarg = torch.arange(1, batch_size * 5 + 1).reshape(batch_size, 5)
        sample_kwargs = {
            "key1": sample_tensor_kwarg,
            "key2": "Not split",
        }

        chunker = ldp.nn.TensorChunker(num_chunks=num_chunks)
        split_args, split_kwargs, dummy_chunk_flags = chunker.chunkify(
            sample_tensor, **sample_kwargs
        )

        assert len(split_args) == num_chunks
        assert len(split_kwargs) == num_chunks
        assert dummy_chunk_flags == [False, False, True]
        assert torch.equal(split_args[0][0], sample_tensor[:1])
        assert torch.equal(split_args[1][0], sample_tensor[1:2])
        assert torch.equal(split_args[2][0], sample_tensor[:1])
        assert torch.equal(split_kwargs[0]["key1"], sample_tensor_kwarg[:1])
        assert torch.equal(split_kwargs[1]["key1"], sample_tensor_kwarg[1:2])
        assert torch.equal(split_kwargs[2]["key1"], sample_tensor_kwarg[:1])
        assert all(split_kwargs[i]["key2"] == "Not split" for i in range(num_chunks))

    def test_dechunkify(self):
        batch_size = 2
        num_chunks = 3

        # Simulate outputs from handlers
        outputs_list = []
        for i in range(num_chunks):
            sequences = torch.tensor([[2 * i], [2 * i + 1]], dtype=torch.long)
            scores = None  # Simplify the test by not including scores
            output = GenerateDecoderOnlyOutput(sequences=sequences, scores=scores)  # type: ignore[arg-type]
            outputs_list.append(output)
        dummy_chunk_flags = [i >= batch_size for i in range(num_chunks)]

        # Dechunkify outputs
        combined_output = ldp.nn.TensorChunker.dechunkify(
            outputs_list, dummy_chunk_flags
        )

        # Since the last chunk was a dummy, it should be excluded
        expected_sequences = [
            torch.tensor([0]),
            torch.tensor([1]),
            torch.tensor([2]),
            torch.tensor([3]),
        ]  # Sequences from first two outputs
        assert all(
            starmap(
                torch.equal,
                zip(combined_output.sequences, expected_sequences, strict=True),
            )
        )
        assert combined_output.scores is None


def randomize_port(
    config: ldp.nn.ParallelModeConfig | None,
) -> ldp.nn.ParallelModeConfig | None:
    """If a test fails and is retried, reset the port to be safe."""
    if not config:
        return None
    return config.model_copy(update={"torch_port": get_unused_port()})


class TestHandlers:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("parallel_mode_config", PARALLEL_MODE_CONFIGS)
    @pytest.mark.usefixtures("seed_zero")
    async def test_generation(
        self,
        parallel_mode_config: ldp.nn.ParallelModeConfig | None,
    ) -> None:
        parallel_mode_config = randomize_port(parallel_mode_config)
        model_config = ldp.nn.LMConfig(
            model="gpt2",
            dtype=ldp.nn.TorchDType.bf16,
            # For this bug: https://github.com/huggingface/transformers/issues/31884
            # Only need to be careful in this case b/c we are checking the exact number
            # of tokens generated.
            tokenizer_args={"clean_up_tokenization_spaces": False},
        )
        # Need this for the tokenizer. TODO: see if we can remove it.
        model_config.resolve_model_location()
        tokenizer = model_config.get_tokenizer()

        handler = ldp.nn.TransformerHandlerConfig(
            lm_config=model_config,
            lm_type=ldp.nn.LMType.GENERATION,
            batch_size=4,
            module_call_fn=ldp.nn.AsyncTransformerInterface.model_generate,
            collate_fn=partial(
                ldp.nn.collate_fn_transformer_left_pad,
                pad_token_id=tokenizer.pad_token_id,
            ),
            decollate_fn=ldp.nn.decollate_fn_transformer_decoder,
            parallel_mode_config=parallel_mode_config,
        ).make_async_module()

        # Set large to encourage generation of end token and test uneven generations.
        max_new_tokens = 100
        outputs = await asyncio.gather(*[
            handler(
                in_text,
                temperature=1.0,
                max_new_tokens=max_new_tokens,
                output_scores=True,
                return_dict_in_generate=True,
                do_sample=True,
            )
            for in_text in ("hello, ", "goodbye, ")
        ])

        for out_text, out_logits in outputs:
            # The length of out_logits must be:
            # * <= max_new_tokens (cannot exceed the max generation length).
            # * >= len(tokenizer(out_text)["input_ids"]).
            # The reason for >= and not == in the second condition is for rare
            # cases where tokenizer optimizes/shrinks tokens as we're re-tokenizing model outputs
            # in the below code.
            assert len(out_logits) <= max_new_tokens
            assert len(out_logits) >= len(tokenizer(out_text)["input_ids"])

    @pytest.mark.parametrize("parallel_mode_config", PARALLEL_MODE_CONFIGS)
    def test_state_dicts(
        self,
        parallel_mode_config: ldp.nn.ParallelModeConfig | None,
    ) -> None:
        parallel_mode_config = randomize_port(parallel_mode_config)
        model_config = ldp.nn.LMConfig(model="gpt2", dtype=ldp.nn.TorchDType.bf16)
        model_config.resolve_model_location()

        handler = ldp.nn.TransformerHandlerConfig(
            lm_config=model_config,
            lm_type=ldp.nn.LMType.GENERATION,
            batch_size=1,
            module_call_fn=ldp.nn.AsyncTransformerInterface.model_generate,
            collate_fn=partial(
                ldp.nn.collate_fn_transformer_left_pad,
                pad_token_id=model_config.get_tokenizer().pad_token_id,
            ),
            decollate_fn=ldp.nn.decollate_fn_transformer_decoder,
            parallel_mode_config=parallel_mode_config,
        ).make_async_module()

        orig_state_dict = handler.state_dict()
        zeros_state_dict = {k: torch.zeros_like(v) for k, v in orig_state_dict.items()}

        if not parallel_mode_config:
            # load_state_dict() is available
            handler.load_state_dict(zeros_state_dict)

        else:
            with pytest.raises(NotImplementedError):
                # check that we can't do this
                handler.load_state_dict(zeros_state_dict)

            # now check we can at least load from disk
            tempdir = Path(mkdtemp())
            tempfile = tempdir / "pytorch_model_fsdp.bin"
            torch.save(zeros_state_dict, tempfile)
            handler.load_checkpoint(tempdir)
            rmtree(tempdir)

        for k, v in handler.state_dict().items():
            assert torch.equal(v, zeros_state_dict[k])

    @pytest.mark.skipif(not TEST_GPUS, reason="Requires GPUs")
    @pytest.mark.parametrize("sharded", [True, False])
    def test_distributed_checkpoints(self, sharded: bool) -> None:
        model_config = ldp.nn.LMConfig(model="gpt2", dtype=ldp.nn.TorchDType.bf16)
        model_config.resolve_model_location()

        handler_config = ldp.nn.TransformerHandlerConfig(
            lm_config=model_config,
            lm_type=ldp.nn.LMType.GENERATION,
            batch_size=1,
            module_call_fn=ldp.nn.AsyncTransformerInterface.model_generate,
            collate_fn=partial(
                ldp.nn.collate_fn_transformer_left_pad,
                pad_token_id=model_config.get_tokenizer().pad_token_id,
            ),
            decollate_fn=ldp.nn.decollate_fn_transformer_decoder,
            parallel_mode_config=ldp.nn.ParallelModeConfig(
                num_workers=2,
                num_cpus_per_worker=1,
                state_dict_type=(
                    StateDictType.SHARDED_STATE_DICT
                    if sharded
                    else StateDictType.FULL_STATE_DICT
                ),
            ),
        )
        handler = ldp.nn.ParallelAsyncTransformer(handler_config)

        # CPU offloading before initializing the optimizer appears to be causing reshapes
        orig_state_dict = handler.state_dict(offload_to_cpu=False)

        # update the model
        @handler.wrap_func(worker_agg_fn=lambda x: sum(x) / len(x))
        def update_model(handler: ldp.nn.TransformerHandler) -> float:
            parameters = list(handler.module.parameters())
            opt = SGD(parameters, lr=0.1)
            x = torch.ones((2, 2), dtype=torch.long).to(handler.module.device)  # type: ignore[arg-type]
            loss = handler.module(x, labels=x).loss
            loss.backward()
            opt.step()
            return loss.item()

        update_model()
        updated_state_dict = handler.state_dict()
        # make sure at least one parameter has changed
        for k, v in updated_state_dict.items():
            if not torch.equal(v, orig_state_dict[k]):
                break
        else:
            raise ValueError("No parameter changed")

        ckpt_path = Path(mkdtemp())
        handler.save_checkpoint(ckpt_path)
        if sharded:
            # one per gpu
            assert (
                len(list((ckpt_path / "pytorch_model_fsdp_0").rglob("*.distcp"))) == 2
            )
        else:
            assert len(list(ckpt_path.rglob("*.bin"))) == 1  # should've been merged

        handler.teardown()

        # Make sure we can load the checkpoint
        new_handler = handler_config.model_copy(
            update={
                "checkpoint": ckpt_path,
                "parallel_mode_config.torch_port": get_unused_port(),
            }
        ).make_async_module()

        # finally, check we got updated_state_dict back
        assert all(
            torch.equal(v, updated_state_dict[k])
            for k, v in new_handler.state_dict().items()
        )

        # cleanup
        rmtree(ckpt_path)

    @pytest.mark.skipif(not TEST_GPUS, reason="Requires GPUs")
    def test_consistent_weights(self):
        # With cpu_ram_efficient_loading, we are asking HF to load the weights on one device
        # and then distribute them. Check that this gives us the same weights as the naive method.
        parallel_configs = {
            "single": None,
            "parallel": ldp.nn.ParallelModeConfig(
                num_workers=2, num_cpus_per_worker=1, cpu_ram_efficient_loading=True
            ),
        }

        model_config = ldp.nn.LMConfig(model="gpt2", dtype=ldp.nn.TorchDType.bf16)
        model_config.resolve_model_location()
        state_dicts = {}

        for mode, parallel_config in parallel_configs.items():
            handler_config = ldp.nn.TransformerHandlerConfig(
                lm_config=model_config,
                lm_type=ldp.nn.LMType.GENERATION,
                batch_size=1,
                module_call_fn=ldp.nn.AsyncTransformerInterface.model_generate,
                collate_fn=partial(
                    ldp.nn.collate_fn_transformer_left_pad,
                    pad_token_id=model_config.get_tokenizer().pad_token_id,
                ),
                decollate_fn=ldp.nn.decollate_fn_transformer_decoder,
                parallel_mode_config=parallel_config,
            )

            handler = handler_config.make_async_module()
            state_dicts[mode] = {k: v.cpu() for k, v in handler.state_dict().items()}

        for k, v in state_dicts["parallel"].items():
            assert torch.equal(v, state_dicts["single"][k])
